//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Wishbone interface for FFT module
//
// Real part of data transferred in wb_data_i[15:0]
// Imaginary part of data tranferred in wb_data_i[31:16]
//
`include "wbhost_parms.v"

module wb_fft (
  clk_i, fft_clk_i, rst_i,
  // Wishbone interface 
  wb_adr_i, wb_data_i, wb_we_i, wb_sel_i, wb_stb_i, wb_ack_i, wb_cyc_i,
  wb_data_o, wb_adr_o, wb_we_o, wb_sel_o, wb_stb_o, wb_cyc_o, wb_ack_o,
  // FFT core interface
  idx_i, conv_rdy_i, re_i, im_i, adr_i,
  re_data_o, re2_data_o, im_data_o, im2_data_o, conv_start_o, dvalid_o, data_clk_o,
  rd_o
  );
  input clk_i;                // Input clock signal
  input fft_clk_i;            // Input clock for FFT core
  input rst_i;                // Input reset signal
  // Wishbone interface
  input [31:0] wb_adr_i;
  input [31:0] wb_data_i;
  input wb_we_i;
  input [3:0] wb_sel_i; 
  input wb_stb_i;
  input wb_ack_i;
  input wb_cyc_i;
  output [31:0] wb_data_o;
  output [31:0] wb_adr_o;
  output wb_we_o;
  output [3:0] wb_sel_o;
  output wb_stb_o;
  output wb_cyc_o;
  output wb_ack_o;
  // FFT core interface
  input [5:0] idx_i;                // Index of dataset to be sent to the fft_top module
  input conv_rdy_i;                 // FFT converted data is ready
  input [17:0] re_i;                // Real part of FFT conversion result
  input [17:0] im_i;                // Imaginary part of FFT conversion result
  input [6:0] adr_i;                // Address of the data being read
  output [13:0] re_data_o;          // Real part of data
  output [13:0] re2_data_o;         // Second output for real data
  output [13:0] im_data_o;          // Imaginary part of data
  output [13:0] im2_data_o;         // Second output  for imaginary part 
  output conv_start_o;              // Start conversion signal - for global syncronization
  output dvalid_o;                  // Signal to clock data
  output data_clk_o;                // Clock for FFT core
  output rd_o;                      // Request to read one word of the result of FFT
  
  
  reg dvalid_o;
  reg conv_start_o;
  reg [31:0] wb_data_o;
  reg [31:0] wb_adr_o = {`BASE_UART,12'h0};    // Only communication with UART
  reg wb_we_o;
  reg [3:0] wb_sel_o;
  reg wb_stb_o;
  reg wb_cyc_o;
  reg wb_ack_o;
  reg [13:0] re_data_o;
  reg [13:0] re2_data_o;
  reg [13:0] im_data_o;
  reg [13:0] im2_data_o;
  reg rd_o;
  
  // Internal registers and wires
  reg [13:0] re_buf1 [0:63];         // First buffer to store even items of real part of data
  reg [13:0] re_buf11 [0:63];        // First buffer to store odd items of real part of data
  reg [13:0] re_buf2 [0:63];         // Second buffer to store even items of real part of data
  reg [13:0] re_buf21 [0:63];        // Second buffer to store odd items of real part of data
  reg [13:0] im_buf1 [0:63];         // First buffer to store even items of imaginary part of data
  reg [13:0] im_buf11 [0:63];        // First buffer to store odd items of imaginary part of data
  reg [13:0] im_buf2 [0:63];         // Second buffer to store even items of imaginary part of data
  reg [13:0] im_buf21 [0:63];        // Second buffer to store odd items of imaginary part of data
  reg [6:0] buf_ptr;                  // Buffer pointer
  wire [6:0] buf_ptr_inv;             // Bit reversed pointer
  reg buf_nmr;                        // Buffer number to be filled
  reg fft_data;                       // When set to one block of FFT converted data is available
  reg tx_done;                        // Conversion of one FFT set of data is completed
  reg [1:0] tx_flag;                  // Flag to specify when to assert rd_o
  reg tx;                             // Flag to make sure rd_o is asserted for one clock cycle only
  reg [8:0] data_cnt;                 // Counter of the data tx'd over Wishbone
  
  assign buf_ptr_inv = {buf_ptr[0],buf_ptr[1],buf_ptr[2],buf_ptr[3],buf_ptr[4],buf_ptr[5],buf_ptr[6]};
  assign data_clk_o = fft_clk_i;
 /* 
  //-----------------------Chipscope logic-----------------------------
  wire [35:0] scope_cntr_w;
  
  chipscope_icon_v1_06_a_0 scope_cntr1 (
    .CONTROL0(scope_cntr_w) // INOUT BUS [35:0]
    );

  chipscope_ila_v1_05_a_0 scope_ila1 (    // Sample on rising edge
    .CONTROL(scope_cntr_w), // INOUT BUS [35:0]
    .CLK(fft_clk_i), // IN
    .TRIG0({3'b000,conv_start_o}), // IN BUS [3:0]
    .TRIG1({idx_i,re_data_o}), // IN BUS [19:0]
    .TRIG2({6'h0,re2_data_o}), // IN BUS [19:0]
    .TRIG3({6'h0,im_data_o}), // IN BUS [19:0]
    .TRIG4({6'h0,im2_data_o}), // IN BUS [19:0]
    .TRIG5()                   // IN BUS [3:0]
);
//----------------------End of chipscope logic-----------------------  
*/  
  
  always @ (negedge clk_i)
    begin
      if (!rst_i)
        begin
          conv_start_o <= 1'b0;
          buf_ptr <= 7'b0;
          buf_nmr <= 1'b0;
          dvalid_o <= 1'b0;
          wb_ack_o <= 1'b0;
        end
      else
        begin
          if (wb_cyc_i)
            begin
              dvalid_o <= 1'b1;
              wb_ack_o <= 1'b1;
              buf_ptr <= buf_ptr + 1'b1;
              if (buf_ptr == 7'b1111111)
                begin
                  buf_nmr <= ~buf_nmr;
                  conv_start_o <= 1'b1;
                end 
              if (~buf_nmr)
                begin
                  if (buf_ptr_inv[0])
                    begin
                      re_buf11[buf_ptr_inv[6:1]] <= wb_data_i[13:0];            // Sign extended to 18-bits data
                      im_buf11[buf_ptr_inv[6:1]] <= wb_data_i[29:16];
                    end
                  else
                    begin
                      re_buf1[buf_ptr_inv[6:1]] <= wb_data_i[13:0];
                      im_buf1[buf_ptr_inv[6:1]] <= wb_data_i[29:16];
                    end
                end
              else
                begin
                  if (buf_ptr_inv[0])
                    begin
                      re_buf21[buf_ptr_inv[6:1]] <= wb_data_i[13:0];
                      im_buf21[buf_ptr_inv[6:1]] <= wb_data_i[29:16];
                    end
                  else
                    begin
                      re_buf2[buf_ptr_inv[6:1]] <= wb_data_i[13:0];
                      im_buf2[buf_ptr_inv[6:1]] <= wb_data_i[29:16];
                    end
                end
            end
          else
            begin
              dvalid_o <= 1'b0;
              wb_ack_o <= 1'b0;
              conv_start_o <= 1'b0;
            end              
        end
    end 
    
  always @ (posedge clk_i)
    begin
      if (!rst_i)
        begin
          fft_data <= 1'b0;
        end
      else
        begin
          if (conv_rdy_i)
            fft_data <= 1'b1;
          else if (tx_done)
            fft_data <= 1'b0;  
        end
    end
    
 always @ (posedge clk_i)
   begin
     if (!rst_i)
       begin
         wb_we_o <= 1'b0;
         wb_sel_o <= 4'hF;
         wb_stb_o <= 1'b0;
         wb_cyc_o <= 1'b0;
         tx_done <= 1'b0;
         rd_o <= 1'b0;
         tx_flag <= 2'b00;
         data_cnt <= 9'h0;
       end
     else
       begin
         if (fft_data)
           begin
             if (tx_flag == 2'b00)
               begin
                 tx_flag <= tx_flag + 1'b1;
                 wb_data_o <= {14'h0,re_i};
                 tx <= 1'b1;
               end
             else if (wb_ack_i)
               begin
                 if (tx_flag == 2'b10)
                   tx_flag <= 2'b00;
                 else
                   tx_flag <= tx_flag + 1'b1;
               end
        
             if (tx_flag == 2'b10)
              begin
                wb_data_o <= {14'h0,im_i};
              end
              
            if ((tx_flag == 2'b10) && tx)
              begin
                rd_o <= 1'b1;
                tx <= 1'b0;
              end
            else
              rd_o <= 1'b0;

           end
           
         if ((fft_data) && ~data_cnt[8])
           begin
             wb_cyc_o <= 1'b1;
             wb_stb_o <= 1'b1;
             wb_we_o <= 1'b1;
           end
         else
           begin
             wb_cyc_o <= 1'b0;
             wb_stb_o <= 1'b0;
             wb_we_o <= 1'b0;
           end
           
         if (fft_data && wb_ack_i)
           data_cnt <= data_cnt + 1'b1;
         else if (~fft_data)
           data_cnt <= 9'h0;
         
         if (data_cnt[8])
           tx_done <= 1'b1;
         else
           tx_done <= 1'b0;  
             
       end
   end
    
  always @ (posedge fft_clk_i)
    begin
      if (buf_nmr)
        begin
          re_data_o <= re_buf1[idx_i];
          re2_data_o <= re_buf11[idx_i];
          im_data_o <= im_buf1[idx_i];
          im2_data_o <= im_buf11[idx_i];
        end
      else
        begin
          re_data_o <= re_buf2[idx_i];
          re2_data_o <= re_buf21[idx_i];
          im_data_o <= im_buf2[idx_i];
          im2_data_o <= im_buf21[idx_i];
        end
    end   
  
endmodule