//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// WB host - for this design is interface between UART and FFT/IFFT modules
//
`include "wbhost_parms.v"

module wb_host (
  clk_i, rst_i,
  // WB signals common to all devices
  wb_adr_o, wb_data_o, wb_we_o, wb_sel_o, wb_stb_o, wb_ack_o,
  // Interface signals with UART controller
  uart_adr_i, uart_data_i, uart_we_i, uart_sel_i, uart_stb_i, uart_cyc_i, uart_ack_i,
  uart_cyc_o,
  // Interface signals with FFT WB module
  fft_adr_i, fft_data_i, fft_we_i, fft_sel_i, fft_stb_i, fft_cyc_i, fft_ack_i,
  fft_cyc_o,
  // Test I/O
  test1_o
  );
  input clk_i;                  // Input clock signal
  input rst_i;                  // Input reset signal
  // WB signals common to all devices
  output [31:0] wb_adr_o;
  output [31:0] wb_data_o;
  output wb_we_o;
  output [3:0] wb_sel_o;
  output wb_stb_o;
  output wb_ack_o;
// Interface with UART controller
  input [31:0] uart_adr_i;
  input [31:0] uart_data_i;
  input uart_we_i;
  input [3:0] uart_sel_i;
  input uart_stb_i;
  input uart_cyc_i;
  input uart_ack_i;
  output uart_cyc_o;
// Interface with FFT module
  input [31:0] fft_adr_i;
  input [31:0] fft_data_i;
  input fft_we_i;
  input [3:0] fft_sel_i;
  input fft_stb_i;
  input fft_cyc_i;
  input fft_ack_i;
  output fft_cyc_o;
// Test I/O
  output [3:0] test1_o;
  
//-----------------WB host state machine--------------------------------
reg [3:0] wbhost_fsm_state;
reg [3:0] wbhost_fsm_next_state;

assign test1_o = wbhost_fsm_state;

always @ (wbhost_fsm_state or uart_cyc_i or fft_cyc_i) // All signals in the sensitivity list change on the posedge
  begin
    wbhost_fsm_next_state = `WBHOST_IDLE;
    case (wbhost_fsm_state)       // Synthesis full case
      `WBHOST_IDLE:
        begin
          if (uart_cyc_i)
            wbhost_fsm_next_state = `WBHOST_UART;
          else if (fft_cyc_i)
            wbhost_fsm_next_state = `WBHOST_FFT;
          else
            wbhost_fsm_next_state = `WBHOST_IDLE;
        end
      `WBHOST_UART:
        begin
          if (uart_cyc_i)
            wbhost_fsm_next_state = `WBHOST_UART;
          else
            wbhost_fsm_next_state = `WBHOST_IDLE;
        end
      `WBHOST_FFT:
        begin
          if (fft_cyc_i)
            wbhost_fsm_next_state = `WBHOST_FFT;
          else
            wbhost_fsm_next_state = `WBHOST_IDLE;
        end
    endcase
  end

always @ (posedge clk_i)
  begin
    if (!rst_i)
      begin
        wbhost_fsm_state <= `WBHOST_IDLE;
      end
    else
      begin
        wbhost_fsm_state <= wbhost_fsm_next_state;
      end
  end
//----------------End of WB host state machine--------------------------
  
  assign wb_adr_o = (wbhost_fsm_state == `WBHOST_UART) ? uart_adr_i :
                    (wbhost_fsm_state == `WBHOST_FFT) ? fft_adr_i : 32'hZZZZZZZZ;
  assign wb_data_o = (wbhost_fsm_state == `WBHOST_UART) ? uart_data_i :
                     (wbhost_fsm_state == `WBHOST_FFT) ? fft_data_i : 32'hZZZZZZZZ;
  assign wb_we_o = (wbhost_fsm_state == `WBHOST_UART) ? uart_we_i :
                   (wbhost_fsm_state == `WBHOST_FFT) ? fft_we_i : 1'b0;
  assign wb_sel_o = (wbhost_fsm_state == `WBHOST_UART) ? uart_sel_i :
                    (wbhost_fsm_state == `WBHOST_FFT) ? fft_sel_i : 4'b0000;
  assign wb_stb_o = (wbhost_fsm_state == `WBHOST_UART) ? uart_stb_i :
                    (wbhost_fsm_state == `WBHOST_FFT) ? fft_stb_i : 1'b0;
  assign wb_cyc_drv = (wbhost_fsm_state == `WBHOST_UART) ? uart_cyc_i :
                      (wbhost_fsm_state == `WBHOST_FFT) ? fft_cyc_i : 1'b0;                     
  assign wb_ack_o = (wb_adr_o[31:12] === `BASE_UART) ? uart_ack_i :
                    (wb_adr_o[31:12] === `BASE_FFT) ? fft_ack_i : 1'b0;                    
  assign uart_cyc_o = (wb_adr_o[31:12] == `BASE_UART) ? wb_cyc_drv : 1'b0;
  assign fft_cyc_o = (wb_adr_o[31:12] == `BASE_FFT) ? wb_cyc_drv : 1'b0; 

endmodule