//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
`include "parms.v"

module mult (
  clk_i, rst_i, re1_i, re2_i, im1_i, im2_i, twr_i, twi_i, start_i, test1_i,
  re1_o, im1_o, re2_o, im2_o, ready_o, first_o, idle_o, test1_o
  );
  
  input clk_i;                    // Input clock
  input rst_i;                    // Input reset signal
  input [17:0] re1_i;             // Input real data - clocked on the negedge
  input [17:0] re2_i;             // Input real data
  input [17:0] im1_i;             // Input imaginary data
  input [17:0] im2_i;             // Input imaginary data
  input [17:0] twr_i;             // Input real part of twiddle - data is clocked on the negedge
  input [17:0] twi_i;             // Input imaginary part of twiddle
  input start_i;                  // Signal to start multiplication - changes on the posedge
  input [3:0] test1_i;                  // Test input
  output [17:0] re1_o;            // Output real part of first result
  output [17:0] im1_o;            // Output imaginary part of first result
  output [17:0] re2_o;            // Output real part of second result
  output [17:0] im2_o;            // Output imaginary part of second result
  output ready_o;                 // Full multiplication cycle is done
  output first_o;                 // Signal to record first result
  output idle_o;                  // FSM is in idle stage
  output [3:0] test1_o;           // Test output
  
  reg ready_o;
  reg [17:0] re1_o;
  reg [17:0] im1_o;
  reg [17:0] re2_o;
  reg [17:0] im2_o;
  
  // Internal registers and wires
  
  wire [47:0] pcout2_w;
  wire [47:0] pcout4_w;
  
  wire [29:0] dre1_w;        // Dummy wires to connect to upper 30 bits of of 48-bits output
  wire [29:0] dim1_w;
  wire [29:0] dre2_w;
  wire [29:0] dim2_w;
  
  wire rst_w = ~rst_i;            // Reset wire for multipliers
  reg [5:0] cnt1;
 
  wire [47:0] ptest_w;
  wire [47:0] ptest2_w;
  
 // Initial values for Opmode registers
 // opmode[1:0], opmode[3:2] and opmode[7] values are changing depending on the FSM state
  reg [7:0] opmode1 = 8'b10000101;        // Opmode register for first multiplier
                            // opmode[1:0] = 2'b01; - Use the multiplier product as X input of postadder
                            // opmode[3:2] - Signal at Z input of postadder;
                            //             = 2'b01 for mult1 and mult3 - Use PCIN
                            //             = 2'b00 for mult2 and mult4 - use zero (propagate multiplier product to P)
                            // opmode[5] = 1'b0 - Value force on the input of the carry-in register (CYI)
                            // opmode[7] - Specifies addition/subtraction for post adder:
                            // for mult1 - subtraction (opmode[7] = 1'b1), all the rest - addition (opmode[7] = 1'b0)
  reg [7:0] opmode2 = 8'b00000001;
  reg [7:0] opmode3 = 8'b00000101;
  reg [7:0] opmode4 = 8'b00000001;
  reg [3:0] cea_r;
  reg [3:0] ceb_r;
  reg [3:0] cep_r;
  reg first_r;
  reg idle_r;
  
  wire [17:0] re2_w;
  wire [17:0] im2_w;
  reg [3:0] mult_fsm_state;
  reg [3:0] mult_fsm_next_state;
  
  assign test1_o = mult_fsm_state;
  assign first_o = first_r;
  assign idle_o = idle_r;
  //----------------------------Internal FSM---------------------------------
  
  always @ (mult_fsm_state or start_i)
    begin
      mult_fsm_next_state = `MULT_IDLE;
      cea_r = 4'h0;
      ceb_r = 4'h0;
      cep_r = 4'hF;
      ready_o = 1'b0;
      first_r = 1'b0;
      idle_r = 1'b0;
      case (mult_fsm_state)                    // Synthesis full case
        `MULT_IDLE:
          begin
            cep_r = 4'h0;
            idle_r = 1'b1;
            if (start_i)
              mult_fsm_next_state = `MULT_M1;
            else
              mult_fsm_next_state = `MULT_IDLE;
          end
        `MULT_M1:        // Multiply second input and twiddle complex numbers 
          begin
            mult_fsm_next_state = `MULT_P1;
            cea_r = 4'hF;
            ceb_r = 4'hF;
          end
        `MULT_P1:        // Propagate mult result to output P for mult2 and mult4
          begin
            mult_fsm_next_state = `MULT_A1;
          end
        `MULT_A1:    // Get final result of complex numbers multiplication - add/subtract mult products
          begin
            mult_fsm_next_state = `MULT_S1;
          end
        `MULT_S1:    // Need to perform addition/subtraction part of FFT algorithm: (1_i + 2_i*tw_i) and
          begin      // (1_i - 2_i*tw_i)
                     // re1_o = 1r_i + (2_i*tw_i)_r - real part of operation
                     // re2_o = 1r_i - (2_i*tw_i)_r
                     // im1_o = 1i_i + (2_i*tw_i)_i - imaginary part of operation
                     // im2_o = 1i_i - (2_i*tw_i)_i
            first_r = 1'b1;
            mult_fsm_next_state = `MULT_DONE;
          end
        `MULT_DONE:
          begin
            mult_fsm_next_state = `MULT_IDLE;
            cep_r = 4'h0;
            ready_o = 1'b1;
          end
      endcase
    end
    
  always @ (posedge clk_i)
    begin
      if (!rst_i)
        begin
          mult_fsm_state <= `MULT_IDLE;
        end
      else
        begin
          mult_fsm_state <= mult_fsm_next_state;
        end
    end
  
  //---------------------------End of internal FSM---------------------------
  

  //-----------------------Chipscope logic-----------------------------
  /*  wire [35:0] scope_cntr_w;
    
    chipscope_icon_v1_06_a_0 scope_cntr1 (
      .CONTROL0(scope_cntr_w) // INOUT BUS [35:0]
  );
  
    chipscope_ila_v1_05_a_0 scope_ila1 (    // Sample on rising edge
      .CONTROL(scope_cntr_w), // INOUT BUS [35:0]
      .CLK(clk_i), // IN
      .TRIG0({test1_i}), // IN BUS [3:0]
      .TRIG1({12'h0,re1_o,im1_o}), // IN BUS [47:0]
      .TRIG2({12'h0,re2_o,im2_o}), // IN BUS [47:0]
      .TRIG3({dim2_w[29:9],im2_w,dim2_w[8:0]}), // IN BUS [47:0]
      .TRIG4({12'h0,re1_i,im1_i}), // IN BUS [47:0]
      .TRIG5({re2_i}), // IN BUS [19:0]
      .TRIG6({im2_i}), // IN BUS [19:0]
      .TRIG7({twr_i}), // IN BUS [19:0]
      .TRIG8({twi_i}), // IN BUS [19:0]
      .TRIG9(mult_fsm_state)               // IN BUS [3:0]
  ); */
  //----------------------End of chipscope logic-----------------------
  
 
    DSP48A1 #(
     .A0REG(0),              // First stage A input pipeline register (0/1)
     .A1REG(1),              // Second stage A input pipeline register (0/1)
     .B0REG(0),              // First stage B input pipeline register (0/1)
     .B1REG(1),              // Second stage B input pipeline register (0/1)
     .CARRYINREG(1),         // CARRYIN input pipeline register (0/1)
     .CARRYINSEL("OPMODE5"), // Specify carry-in source, "CARRYIN" or "OPMODE5" 
     .CARRYOUTREG(1),        // CARRYOUT output pipeline register (0/1)
     .CREG(0),               // C input pipeline register (0/1)
     .DREG(0),               // D pre-adder input pipeline register (0/1)
     .MREG(0),               // M pipeline register (0/1)
     .OPMODEREG(1),          // Enable=1/disable=0 OPMODE input pipeline registers
     .PREG(1),               // P output pipeline register (0/1)
     .RSTTYPE("SYNC")        // Specify reset type, "SYNC" or "ASYNC" 
  )
  DSP48A1_mult1 (
     // Cascade Ports: 18-bit (each) output: Ports to cascade from one DSP48 to another
     .BCOUT(BCOUT),           // 18-bit output: B port cascade output
     .PCOUT(),           // 48-bit output: P cascade output (if used, connect to PCIN of another DSP48A1)
     // Data Ports: 1-bit (each) output: Data input and output ports
     .CARRYOUT(CARRYOUT),     // 1-bit output: carry output (if used, connect to CARRYIN pin of another
                              // DSP48A1)

     .CARRYOUTF(CARRYOUTF),   // 1-bit output: fabric carry output
     .M(),                    // 36-bit output: fabric multiplier data output
     .P({dre2_w[29:9],re2_w,dre2_w[8:0]}),                   // 48-bit output: data output
     // Cascade Ports: 48-bit (each) input: Ports to cascade from one DSP48 to another
     .PCIN(pcout2_w),             // 48-bit input: P cascade input (if used, connect to PCOUT of another DSP48A1)
     // Control Input Ports: 1-bit (each) input: Clocking and operation mode
     .CLK(clk_i),               // 1-bit input: clock input
     .OPMODE(opmode1),         // 8-bit input: operation mode input
     // Data Ports: 18-bit (each) input: Data input and output ports
     .A(twi_i),                   // 18-bit input: A data input
     .B(im2_i),                   // 18-bit input: B data input (connected to fabric or BCOUT of adjacent DSP48A1)
     .C(),                   // 48-bit input: C data input
     .CARRYIN(CARRYIN),       // 1-bit input: carry input signal (if used, connect to CARRYOUT pin of another
                              // DSP48A1)

     .D(D),                   // 18-bit input: B pre-adder data input
     // Reset/Clock Enable Input Ports: 1-bit (each) input: Reset and enable input ports
     .CEA(cea_r[0]),               // 1-bit input: active high clock enable input for A registers
     .CEB(ceb_r[0]),               // 1-bit input: active high clock enable input for B registers
     .CEC(1'b0),               // 1-bit input: active high clock enable input for C registers
     .CECARRYIN(1'b1),   // 1-bit input: active high clock enable input for CARRYIN registers
     .CED(1'b0),               // 1-bit input: active high clock enable input for D registers
     .CEM(1'b0),               // 1-bit input: active high clock enable input for multiplier registers
     .CEOPMODE(1'b1),     // 1-bit input: active high clock enable input for OPMODE registers
     .CEP(cep_r[0]),               // 1-bit input: active high clock enable input for P registers
     .RSTA(rst_w),             // 1-bit input: reset input for A pipeline registers
     .RSTB(rst_w),             // 1-bit input: reset input for B pipeline registers
     .RSTC(rst_w),             // 1-bit input: reset input for C pipeline registers
     .RSTCARRYIN(rst_w), // 1-bit input: reset input for CARRYIN pipeline registers
     .RSTD(rst_w),             // 1-bit input: reset input for D pipeline registers
     .RSTM(rst_w),             // 1-bit input: reset input for M pipeline registers
     .RSTOPMODE(rst_w),   // 1-bit input: reset input for OPMODE pipeline registers
     .RSTP(rst_w)              // 1-bit input: reset input for P pipeline registers
  );

  // End of DSP48A1_inst instantiation
  
    
      DSP48A1 #(
       .A0REG(0),              // First stage A input pipeline register (0/1)
       .A1REG(1),              // Second stage A input pipeline register (0/1)
       .B0REG(0),              // First stage B input pipeline register (0/1)
       .B1REG(1),              // Second stage B input pipeline register (0/1)
       .CARRYINREG(1),         // CARRYIN input pipeline register (0/1)
       .CARRYINSEL("OPMODE5"), // Specify carry-in source, "CARRYIN" or "OPMODE5" 
       .CARRYOUTREG(1),        // CARRYOUT output pipeline register (0/1)
       .CREG(0),               // C input pipeline register (0/1)
       .DREG(0),               // D pre-adder input pipeline register (0/1)
       .MREG(0),               // M pipeline register (0/1)
       .OPMODEREG(1),          // Enable=1/disable=0 OPMODE input pipeline registers
       .PREG(1),               // P output pipeline register (0/1)
       .RSTTYPE("SYNC")        // Specify reset type, "SYNC" or "ASYNC" 
    )
    DSP48A1_mult2 (
       // Cascade Ports: 18-bit (each) output: Ports to cascade from one DSP48 to another
       .BCOUT(BCOUT),           // 18-bit output: B port cascade output
       .PCOUT(pcout2_w),           // 48-bit output: P cascade output (if used, connect to PCIN of another DSP48A1)
       // Data Ports: 1-bit (each) output: Data input and output ports
       .CARRYOUT(CARRYOUT),     // 1-bit output: carry output (if used, connect to CARRYIN pin of another
                                // DSP48A1)
  
       .CARRYOUTF(CARRYOUTF),   // 1-bit output: fabric carry output
       .M(),                   // 36-bit output: fabric multiplier data output
       .P(ptest_w),                   // 48-bit output: data output
       // Cascade Ports: 48-bit (each) input: Ports to cascade from one DSP48 to another
       .PCIN(),             // 48-bit input: P cascade input (if used, connect to PCOUT of another DSP48A1)
       // Control Input Ports: 1-bit (each) input: Clocking and operation mode
       .CLK(clk_i),               // 1-bit input: clock input
       .OPMODE(opmode2),         // 8-bit input: operation mode input
       // Data Ports: 18-bit (each) input: Data input and output ports
       .A(twr_i),                   // 18-bit input: A data input
       .B(re2_i),                   // 18-bit input: B data input (connected to fabric or BCOUT of adjacent DSP48A1)
                               // re2_i
       .C(),                   // 48-bit input: C data input
       .CARRYIN(CARRYIN),       // 1-bit input: carry input signal (if used, connect to CARRYOUT pin of another
                                // DSP48A1)
  
       .D(),                   // 18-bit input: B pre-adder data input
       // Reset/Clock Enable Input Ports: 1-bit (each) input: Reset and enable input ports
       .CEA(cea_r[1]),               // 1-bit input: active high clock enable input for A registers
       .CEB(ceb_r[1]),               // 1-bit input: active high clock enable input for B registers
       .CEC(1'b0),               // 1-bit input: active high clock enable input for C registers
       .CECARRYIN(1'b1),   // 1-bit input: active high clock enable input for CARRYIN registers
       .CED(1'b0),               // 1-bit input: active high clock enable input for D registers
       .CEM(1'b0),               // 1-bit input: active high clock enable input for multiplier registers
       .CEOPMODE(1'b1),     // 1-bit input: active high clock enable input for OPMODE registers
       .CEP(cep_r[1]),               // 1-bit input: active high clock enable input for P registers
       .RSTA(rst_w),             // 1-bit input: reset input for A pipeline registers
       .RSTB(rst_w),             // 1-bit input: reset input for B pipeline registers
       .RSTC(rst_w),             // 1-bit input: reset input for C pipeline registers
       .RSTCARRYIN(rst_w), // 1-bit input: reset input for CARRYIN pipeline registers
       .RSTD(rst_w),             // 1-bit input: reset input for D pipeline registers
       .RSTM(rst_w),             // 1-bit input: reset input for M pipeline registers
       .RSTOPMODE(rst_w),   // 1-bit input: reset input for OPMODE pipeline registers
       .RSTP(rst_w)              // 1-bit input: reset input for P pipeline registers
    );
  
    // End of DSP48A1_inst instantiation
    
        
        DSP48A1 #(
         .A0REG(0),              // First stage A input pipeline register (0/1)
         .A1REG(1),              // Second stage A input pipeline register (0/1)
         .B0REG(0),              // First stage B input pipeline register (0/1)
         .B1REG(1),              // Second stage B input pipeline register (0/1)
         .CARRYINREG(1),         // CARRYIN input pipeline register (0/1)
         .CARRYINSEL("OPMODE5"), // Specify carry-in source, "CARRYIN" or "OPMODE5" 
         .CARRYOUTREG(1),        // CARRYOUT output pipeline register (0/1)
         .CREG(0),               // C input pipeline register (0/1)
         .DREG(0),               // D pre-adder input pipeline register (0/1)
         .MREG(0),               // M pipeline register (0/1)
         .OPMODEREG(1),          // Enable=1/disable=0 OPMODE input pipeline registers
         .PREG(1),               // P output pipeline register (0/1)
         .RSTTYPE("SYNC")        // Specify reset type, "SYNC" or "ASYNC" 
      )
      DSP48A1_mult3 (
         // Cascade Ports: 18-bit (each) output: Ports to cascade from one DSP48 to another
         .BCOUT(BCOUT),           // 18-bit output: B port cascade output
         .PCOUT(),           // 48-bit output: P cascade output (if used, connect to PCIN of another DSP48A1)
         // Data Ports: 1-bit (each) output: Data input and output ports
         .CARRYOUT(CARRYOUT),     // 1-bit output: carry output (if used, connect to CARRYIN pin of another
                                  // DSP48A1)
    
         .CARRYOUTF(CARRYOUTF),   // 1-bit output: fabric carry output
         .M(),                   // 36-bit output: fabric multiplier data output
         .P({dim2_w[29:9],im2_w,dim2_w[8:0]}),                   // 48-bit output: data output
         // Cascade Ports: 48-bit (each) input: Ports to cascade from one DSP48 to another
         .PCIN(pcout4_w),             // 48-bit input: P cascade input (if used, connect to PCOUT of another DSP48A1)
         // Control Input Ports: 1-bit (each) input: Clocking and operation mode
         .CLK(clk_i),               // 1-bit input: clock input
         .OPMODE(opmode3),         // 8-bit input: operation mode input
         // Data Ports: 18-bit (each) input: Data input and output ports
         .A(twi_i),                   // 18-bit input: A data input
         .B(re2_i),                   // 18-bit input: B data input (connected to fabric or BCOUT of adjacent DSP48A1)
         .C(),                   // 48-bit input: C data input
         .CARRYIN(CARRYIN),       // 1-bit input: carry input signal (if used, connect to CARRYOUT pin of another
                                  // DSP48A1)
    
         .D(D),                   // 18-bit input: B pre-adder data input
         // Reset/Clock Enable Input Ports: 1-bit (each) input: Reset and enable input ports
         .CEA(cea_r[2]),               // 1-bit input: active high clock enable input for A registers
         .CEB(ceb_r[2]),               // 1-bit input: active high clock enable input for B registers
         .CEC(1'b0),               // 1-bit input: active high clock enable input for C registers
         .CECARRYIN(1'b1),   // 1-bit input: active high clock enable input for CARRYIN registers
         .CED(1'b0),               // 1-bit input: active high clock enable input for D registers
         .CEM(1'b0),               // 1-bit input: active high clock enable input for multiplier registers
         .CEOPMODE(1'b1),     // 1-bit input: active high clock enable input for OPMODE registers
         .CEP(cep_r[2]),               // 1-bit input: active high clock enable input for P registers
         .RSTA(rst_w),             // 1-bit input: reset input for A pipeline registers
         .RSTB(rst_w),             // 1-bit input: reset input for B pipeline registers
         .RSTC(rst_w),             // 1-bit input: reset input for C pipeline registers
         .RSTCARRYIN(rst_w), // 1-bit input: reset input for CARRYIN pipeline registers
         .RSTD(rst_w),             // 1-bit input: reset input for D pipeline registers
         .RSTM(rst_w),             // 1-bit input: reset input for M pipeline registers
         .RSTOPMODE(rst_w),   // 1-bit input: reset input for OPMODE pipeline registers
         .RSTP(rst_w)              // 1-bit input: reset input for P pipeline registers
      );
    
      // End of DSP48A1_inst instantiation
      
            
          DSP48A1 #(
           .A0REG(0),              // First stage A input pipeline register (0/1)
           .A1REG(1),              // Second stage A input pipeline register (0/1)
           .B0REG(0),              // First stage B input pipeline register (0/1)
           .B1REG(1),              // Second stage B input pipeline register (0/1)
           .CARRYINREG(1),         // CARRYIN input pipeline register (0/1)
           .CARRYINSEL("OPMODE5"), // Specify carry-in source, "CARRYIN" or "OPMODE5" 
           .CARRYOUTREG(1),        // CARRYOUT output pipeline register (0/1)
           .CREG(0),               // C input pipeline register (0/1)
           .DREG(0),               // D pre-adder input pipeline register (0/1)
           .MREG(0),               // M pipeline register (0/1)
           .OPMODEREG(1),          // Enable=1/disable=0 OPMODE input pipeline registers
           .PREG(1),               // P output pipeline register (0/1)
           .RSTTYPE("SYNC")        // Specify reset type, "SYNC" or "ASYNC" 
        )
        DSP48A1_mult4 (
           // Cascade Ports: 18-bit (each) output: Ports to cascade from one DSP48 to another
           .BCOUT(BCOUT),           // 18-bit output: B port cascade output
           .PCOUT(pcout4_w),           // 48-bit output: P cascade output (if used, connect to PCIN of another DSP48A1)
           // Data Ports: 1-bit (each) output: Data input and output ports
           .CARRYOUT(CARRYOUT),     // 1-bit output: carry output (if used, connect to CARRYIN pin of another
                                    // DSP48A1)
      
           .CARRYOUTF(CARRYOUTF),   // 1-bit output: fabric carry output
           .M(),                   // 36-bit output: fabric multiplier data output
           .P(ptest2_w),                   // 48-bit output: data output
           // Cascade Ports: 48-bit (each) input: Ports to cascade from one DSP48 to another
           .PCIN(),             // 48-bit input: P cascade input (if used, connect to PCOUT of another DSP48A1)
           // Control Input Ports: 1-bit (each) input: Clocking and operation mode
           .CLK(clk_i),               // 1-bit input: clock input
           .OPMODE(opmode4),         // 8-bit input: operation mode input
           // Data Ports: 18-bit (each) input: Data input and output ports
           .A(twr_i),                   // 18-bit input: A data input
           .B(im2_i),                   // 18-bit input: B data input (connected to fabric or BCOUT of adjacent DSP48A1)
           .C(),                   // 48-bit input: C data input
           .CARRYIN(CARRYIN),       // 1-bit input: carry input signal (if used, connect to CARRYOUT pin of another
                                    // DSP48A1)
      
           .D(),                   // 18-bit input: B pre-adder data input
           // Reset/Clock Enable Input Ports: 1-bit (each) input: Reset and enable input ports
           .CEA(cea_r[3]),               // 1-bit input: active high clock enable input for A registers
           .CEB(ceb_r[3]),               // 1-bit input: active high clock enable input for B registers
           .CEC(1'b0),               // 1-bit input: active high clock enable input for C registers
           .CECARRYIN(1'b1),   // 1-bit input: active high clock enable input for CARRYIN registers
           .CED(1'b0),               // 1-bit input: active high clock enable input for D registers
           .CEM(1'b0),               // 1-bit input: active high clock enable input for multiplier registers
           .CEOPMODE(1'b1),     // 1-bit input: active high clock enable input for OPMODE registers
           .CEP(cep_r[3]),               // 1-bit input: active high clock enable input for P registers
           .RSTA(rst_w),             // 1-bit input: reset input for A pipeline registers
           .RSTB(rst_w),             // 1-bit input: reset input for B pipeline registers
           .RSTC(rst_w),             // 1-bit input: reset input for C pipeline registers
           .RSTCARRYIN(rst_w), // 1-bit input: reset input for CARRYIN pipeline registers
           .RSTD(rst_w),             // 1-bit input: reset input for D pipeline registers
           .RSTM(rst_w),             // 1-bit input: reset input for M pipeline registers
           .RSTOPMODE(rst_w),   // 1-bit input: reset input for OPMODE pipeline registers
           .RSTP(rst_w)              // 1-bit input: reset input for P pipeline registers
        );
      
        // End of DSP48A1_inst instantiation
        
  always @ (posedge clk_i)
    begin
      if (!rst_i)
        begin
          cnt1 <= 6'h0;
        end
      else
        begin
          re1_o <= re1_i + re2_w;
          re2_o <= re1_i - re2_w;
          im1_o <= im1_i + im2_w;
          im2_o <= im1_i - im2_w;
        end
    end
endmodule

  